"""
The `preconstruct` module creates a dataset that is arranged to facilitate
neural decoding, i.e. reconstructing a stimulus given the neural
response that it induces.

To create the dataset, you must create a `preconstruct.sources.DataSource`
and pass it a `preconstruct.dataset.DatasetBuilder` and configure your dataset.

## What this library can do

- Download your [pprox](https://meliza.org/spec:2/pprox/) files
and stimuli in WAVE format from [Neurobank](https://github.com/melizalab/neurobank)
(`preconstruct.sources.NeurobankSource`)
- Allow you to easily switch between different sources of data, such as
local files (`preconstruct.sources.FsSource`)
or Python dictionaries (`preconstruct.sources.MemorySource`)
- Create gammatone-based spectrogram from raw WAVE data of the stimuli
(`preconstruct.dataset.DatasetBuilder.add_stimuli`)
- Automatically update your pprox files to match the
[stimtrial specification](https://github.com/melizalab/lab_specs/blob/main/specs/stimulus_trial.md)
and store the data in a convenient pandas DataFrame
- Convert the point process from the pprox data into an array of fixed duration
bins containing the count of spikes occuring within the corresponding time window
(`preconstruct.dataset.DatasetBuilder.bin_responses`)
- Project the spikes into an alternate basis (`preconstruct.basisfunctions`)
- Rearrange the binned pprox data into windows of a given duration `tau` for
input into a neural decoding model
"""
from joblib import Memory
from appdirs import user_cache_dir

from pathlib import Path
from single_source import get_version

__version__ = get_version(__name__, Path(__file__).parent.parent)

APP_NAME = "preconstruct"
APP_AUTHOR = "melizalab"


_cache_dir = user_cache_dir(APP_NAME, APP_AUTHOR)
_mem = Memory(_cache_dir, verbose=0)

from preconstruct.dataset import Dataset, DatasetBuilder
from preconstruct import basisfunctions



import pytest
import numpy as np
from preconstruct.sources import NeurobankSource, MemorySource


@pytest.fixture
def stimtrial_pprox():
    return {
        "neuron_1": {
            "$schema": "https://meliza.org/spec:2/stimtrial.json#",
            "pprox": [
                {
                    "events": [0.2, 1, 1.5],
                    "interval": [0, 3],
                    "stimulus": {
                        "interval": [1, 2],
                        "name": "song_1",
                    },
                },
            ],
        }
    }


@pytest.fixture
def cn_pprox():
    return {
        "neuron_1": {
            "protocol": "songs",
            "entry_metadata": {
                "sampling_rate": 10,
            },
            "pprox": [
                {
                    "index": 0,
                    "events": [0.2, 1, 1.5],
                    "recording": {
                        "start": 0,
                        "stop": 30,
                    },
                    "stim": "song_1",
                    "stim_on": 1,
                }
            ],
        }
    }


@pytest.fixture
def ar_pprox():
    return {
        "neuron_1": {
            "experiment": "induction",
            "pprox": [
                {
                    "trial": 0,
                    "units": "ms",
                    "event": [200, 1000, 1500],
                    "stim_uuid": "song_1",
                    "stimulus": "song",
                    "condition": 1,
                    "stim_on": 1,
                }
            ],
        }
    }


@pytest.fixture
def stimuli():
    sample_rate = 44100
    samples = np.random.normal(0, 1000, 44100)
    return {"song_1": (sample_rate, samples)}


@pytest.fixture
def mem_data_source(stimtrial_pprox, stimuli):
    responses = stimtrial_pprox
    return MemorySource(responses, stimuli)


@pytest.fixture
async def real_data_source():
    responses = ["P120_1_1_c92", "P120_1_1_c89"]
    stimuli = [
        "c95zqjxq",
        "g29wxi4q",
        "igmi8fxa",
        "jkexyrd5",
        "l1a3ltpy",
        "mrel2o09",
        "p1mrfhop",
        "vekibwgj",
        "w08e1crn",
        "ztqee46x",
    ]
    url = "https://gracula.psyc.virginia.edu/neurobank/"
    return await NeurobankSource.create(url, stimuli, responses)


@pytest.fixture
def known_spectrogram():
    fs = 10e3
    N = 1e4
    amp = 2 * np.sqrt(2)
    time = np.arange(N) / float(fs)
    freq = 3e3
    x = amp * np.sin(2 * np.pi * freq * time)
    return {"song_1": (fs, x)}


@pytest.fixture
def known_spectrogram_source(stimtrial_pprox, known_spectrogram):
    responses = stimtrial_pprox
    return MemorySource(responses, known_spectrogram)



"""
Classes that allow you to configure different ways of representing stimuli

Only concrete classes (not abstract) can be used as input to a
`preconstruct.dataset.DatasetBuilder`
"""
from abc import ABC, abstractmethod
from typing import Optional, Sequence, Tuple, Mapping, Callable

import numpy as np
import pandas as pd
from scipy import signal
from gammatone.gtgram import gtgram
from gammatone.filters import centre_freqs

from preconstruct.sources import Wav, DataSource
from preconstruct import _mem


class StimuliFormat(ABC):
    """Abstract base class for representing stimuli"""

    @abstractmethod
    def format_from_wav(
        self, name: str, wav_data: Wav, interval: Tuple[float, float], time_step: float
    ) -> pd.DataFrame:
        """return formatted version of WAVE data"""

    def to_values(self, stim_df) -> np.ndarray:
        """convert from pd.DataFrame to ndarray"""
        return stim_df.values

    def create_dataframe(
        self,
        data_source: DataSource,
        time_step: float,
        intervals: pd.Series,
    ) -> pd.DataFrame:
        """build `stimuli` DataFrame"""
        wav_data = data_source.get_stimuli()
        stimuli = (
            pd.DataFrame(wav_data, index=["sample_rate", "samples"])
            .T.rename_axis("stimulus.name")
            .join(intervals.rename("interval"), how="right")
        )
        formatted_stimuli = pd.concat(
            {
                index: self.format_from_wav(
                    index,
                    (data["sample_rate"], data["samples"]),
                    data["interval"],
                    time_step,
                )
                for index, data in stimuli.iterrows()
            }
        ).fillna(0)
        return formatted_stimuli


class LogTransformable(StimuliFormat):
    """Abstract base class for representing stimuli that can be log transformed

    provides the log_transform_compress keyword argument
    """

    def __init__(self, log_transform_compress: Optional[float] = None, **_kwargs):
        self.compress = log_transform_compress

    def format_from_wav(self, *args) -> pd.DataFrame:
        spectrogram = self._raw_format_from_wav(*args)
        if self.compress is not None:
            return np.log10(spectrogram + self.compress) - np.log10(self.compress)
        return spectrogram

    @abstractmethod
    def _raw_format_from_wav(
        self, wav_data: Wav, interval: Tuple[float, float], time_step: float
    ) -> pd.DataFrame:
        """format_from_wav without log_transform applied"""


class Spectrogram(LogTransformable):
    """Spectrogram format

    Consult the [scipy documentation for spectrogram](https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.spectrogram.html)
    to see a list of acceptable keywords. (`x`, `fs`, and `nperseg` will be set automatically.)
    """

    def __init__(
        self,
        min_frequency: Optional[int] = None,
        max_frequency: Optional[int] = None,
        log_transform_compress: Optional[float] = None,
        **kwargs
    ):
        self.kwargs = kwargs
        self.min_frequency = min_frequency
        self.max_frequency = max_frequency
        super().__init__(log_transform_compress, **kwargs)

    def _raw_format_from_wav(
        self, _name: str, wav_data: Wav, interval: Tuple[float, float], time_step: float
    ) -> pd.DataFrame:
        sample_rate, samples = wav_data
        nperseg = int(sample_rate * time_step)
        f, t, spectrogram = signal.spectrogram(
            samples, sample_rate, nperseg=nperseg, **self.kwargs
        )
        start_time, _ = interval
        spectrogram_df = pd.DataFrame(spectrogram.T, columns=f, index=t + start_time)
        if self.min_frequency:
            spectrogram_df = spectrogram_df[
                spectrogram_df.columns[spectrogram_df.columns >= self.min_frequency]
            ]
        if self.max_frequency:
            spectrogram_df = spectrogram_df[
                spectrogram_df.columns[spectrogram_df.columns <= self.max_frequency]
            ]
        return spectrogram_df


class Gammatone(LogTransformable):
    """Gammatone format"""

    def __init__(
        self,
        window_time=0.001,
        frequency_bin_count=50,
        min_frequency=500,
        max_frequency=8000,
        log_transform_compress: Optional[float] = None,
    ):
        super().__init__(log_transform_compress)
        self.params = {
            "window_time": window_time,
            "channels": frequency_bin_count,
            "f_min": min_frequency,
            "f_max": max_frequency,
        }

    def _raw_format_from_wav(
        self, _name, wav_data: Wav, interval: Tuple[float, float], time_step: float
    ) -> pd.DataFrame:
        sample_rate, samples = wav_data
        self.params["hop_time"] = time_step
        spectrogram = _mem.cache(gtgram)(samples, sample_rate, **self.params)
        return pd.DataFrame(
            spectrogram.T,
            columns=np.flip(
                centre_freqs(
                    sample_rate,
                    self.params["channels"],
                    self.params["f_min"],
                    self.params["f_max"],
                )
            ),
            index=np.linspace(*interval, spectrogram.shape[1]),
        )


class SyllableCategorical(StimuliFormat):
    """Each syllable gets its own identifier"""

    def __init__(self, **kwargs) -> None:
        """kwargs passed to scipy.signal.find_peaks"""
        super().__init__()
        self.kwargs = kwargs

    def format_from_wav(
        self, name: str, wav_data: Wav, interval: Tuple[float, float], time_step: float
    ) -> pd.DataFrame:
        """return formatted version of WAVE data"""
        start, stop = interval
        peaks = self.find_syllable_boundaries(wav_data, interval, time_step)
        syllable_boundaries = pd.Series(
            [
                start,
                *peaks,
                stop + time_step,
            ]
        )
        t = np.arange(start, stop + time_step, time_step)
        syllables = (
            pd.cut(
                pd.Index(t).to_series(), bins=syllable_boundaries, include_lowest=True
            )
            .astype("object")
            .apply(lambda x: (name, x))
            .rename("syllable")
        )
        return syllables

    def find_syllable_boundaries(self, wav_data, interval, time_step):
        sample_rate, samples = wav_data
        start, stop = interval
        # we want smaller windows than the size of time_step for identifying syllables
        # with greater precision
        nperseg = int(sample_rate * time_step) // 2
        _, _, spectrogram = signal.spectrogram(samples, sample_rate, nperseg=nperseg)
        power = -np.mean(np.log10(spectrogram + 1), axis=0)
        # syllable boundaries should be moments when the sound gets quiet for a bit
        find_peaks_kwargs = {
            "height": np.median(power),
            "distance": 5,
            "prominence": np.std(power),
        }
        find_peaks_kwargs.update(self.kwargs)
        peaks, _ = signal.find_peaks(power, **find_peaks_kwargs)
        return np.linspace(start, stop, len(power))[peaks]



"""Loading data for neural decoding

## Examples

Let's build a simple dataset for neural decoding. The first step is determining
what data we'll use. Depending on where our data is stored we'll use different
classes that inherit from preconstruct.sources.DataSource. For example, if we want to
use data from a local folder on our computer we might use preconstruct.sources.FsSource.
For other options, see preconstruct.sources. In this example, we'll use
preconstruct.sources.NeurobankSource, which will automatically download a list of
identifiers from Neurobank. Let's suppose we know we want to use a pprox responses
from two files with identifiers `P120_1_1_c92` and `P120_1_1_c89`.
>>> from preconstruct.sources import NeurobankSource
>>> import asyncio
>>> responses = ['P120_1_1_c92', 'P120_1_1_c89']
>>> stimuli = [] # we'll leave this empty for now
>>> url = 'https://gracula.psyc.virginia.edu/neurobank/'
>>> test_source = asyncio.run(NeurobankSource.create(url, stimuli, responses))

We can't use this `DataSource` to build a dataset because it doesn't contain any
of the stimuli that were presented during the recording, but we can easily get
a list of the stimuli identifiers:
>>> stimuli = list(test_source.stimuli_names_from_pprox())
>>> sorted(stimuli)
['c95zqjxq', 'g29wxi4q', 'igmi8fxa', 'jkexyrd5', 'l1a3ltpy', 'mrel2o09', \
'p1mrfhop', 'vekibwgj', 'w08e1crn', 'ztqee46x']

Let's make a new DataSource that includes the stimuli
>>> data_source = asyncio.run(NeurobankSource.create(url, stimuli, responses))

Now we can start building our dataset. We will put our data into a
`DatasetBuilder`, which will allow us to configure the format of our dataset.
>>> from preconstruct.dataset import DatasetBuilder
>>> builder = DatasetBuilder()
>>> builder.set_data_source(data_source)
>>> builder.load_responses()

The first choice we have to make is the size of our time steps. This will
determine the granularity of the time axis for both the spikes and the stimuli.
The unit for this argument and all other time values will be seconds.
>>> builder.bin_responses(time_step=0.005) # 5 ms

Next, we load the stimuli. We must choose parameters to control how the gammatone
spectrograms are generated. Consult DatasetBuilder.add_stimuli for details on each
argument.
>>> from preconstruct.stimuliformats import Gammatone
>>> builder.add_stimuli(
...     Gammatone(
...         window_time=0.005,
...         frequency_bin_count=50,
...         min_frequency=500,
...         max_frequency=8000,
...         log_transform_compress=1,
...     )
... )

Now we will convert the binned spikes into a lagged matrix, with a with a window
of size tau.
>>> builder.create_time_lags(tau=0.3)

Our next step is to combine data from multiple presentations of the same stimulus.
(This is optional.)
>>> builder.pool_trials()

We have finished building our dataset.
We should investigate our object a bit, to make sure we understand how
it's structured.
>>> dataset = builder.get_dataset()

>>> dataset.responses.index
MultiIndex([('c95zqjxq',                1.0),
            ('c95zqjxq', 1.0050000000000001),
            ('c95zqjxq',               1.01),
            ('c95zqjxq', 1.0150000000000001),
            ('c95zqjxq',               1.02),
            ('c95zqjxq',              1.025),
            ('c95zqjxq',               1.03),
            ('c95zqjxq',              1.035),
            ('c95zqjxq',               1.04),
            ('c95zqjxq',              1.045),
            ...
            ('ztqee46x',              3.095),
            ('ztqee46x',                3.1),
            ('ztqee46x',              3.105),
            ('ztqee46x',               3.11),
            ('ztqee46x',              3.115),
            ('ztqee46x',               3.12),
            ('ztqee46x',              3.125),
            ('ztqee46x',               3.13),
            ('ztqee46x', 3.1350000000000002),
            ('ztqee46x',               3.14)],
           names=['stimulus.name', 'time'], length=4112)
>>> dataset.responses.columns
MultiIndex([('P120_1_1_c89',                 0.0),
            ('P120_1_1_c89',               0.005),
            ('P120_1_1_c89',                0.01),
            ('P120_1_1_c89',               0.015),
            ('P120_1_1_c89',                0.02),
            ('P120_1_1_c89',               0.025),
            ('P120_1_1_c89',                0.03),
            ('P120_1_1_c89',               0.035),
            ('P120_1_1_c89',                0.04),
            ('P120_1_1_c89',               0.045),
            ...
            ('P120_1_1_c92',                0.25),
            ('P120_1_1_c92',               0.255),
            ('P120_1_1_c92',                0.26),
            ('P120_1_1_c92',               0.265),
            ('P120_1_1_c92',                0.27),
            ('P120_1_1_c92',               0.275),
            ('P120_1_1_c92',                0.28),
            ('P120_1_1_c92', 0.28500000000000003),
            ('P120_1_1_c92',                0.29),
            ('P120_1_1_c92',               0.295)],
           names=['neuron', 'offset'], length=120)

Let's use our dataset to perform a simple neural decoding task

>>> from sklearn.linear_model import Ridge
>>> import numpy as np
>>> training_stimuli = ['c95zqjxq', 'g29wxi4q', 'igmi8fxa', 'jkexyrd5', 'l1a3ltpy', 'mrel2o09']
>>> test_stimuli = set(dataset.responses.index.get_level_values(level='stimulus.name')).difference(training_stimuli)
>>> X, Y = dataset[list(training_stimuli)]
>>> X.shape, Y.shape
((2476, 120), (2476, 50))
>>> model = Ridge(alpha=1.0)
>>> model.fit(X, Y)
Ridge()
>>> model.score(X, Y)
0.30159992
>>> X_test, Y_test = dataset[list(test_stimuli)]
>>> model.score(X_test, Y_test)
0.15341238

"""
from typing import List, Optional, Dict, Tuple

import numpy as np
import pandas as pd
from scipy.linalg import hankel

from preconstruct.sources import DataSource
from preconstruct.basisfunctions import Basis
from preconstruct.stimuliformats import StimuliFormat


class Dataset:
    """Holds constructed response matrix and stimuli

    [Quick guide to Pandas indexing](https://pandas.pydata.org/docs/getting_started/intro_tutorials/03_subset_data.html#min-tut-03-subset)

    Note: You can index straight into a `Dataset` object to get stimuli and
    responses in numpy array format, but you must select rows rather than
    columns. This corresponds to the `df.loc[]` Pandas syntax, rather than
    `df[]`.

    <!--
    >>> import asyncio
    >>> from preconstruct.sources import NeurobankSource
    >>> from preconstruct.stimuliformats import Gammatone
    >>> responses = ['P120_1_1_c92']
    >>> url = 'https://gracula.psyc.virginia.edu/neurobank/'
    >>> stimuli = ['c95zqjxq', 'g29wxi4q', 'igmi8fxa', 'jkexyrd5', 'l1a3ltpy',
    ...         'mrel2o09', 'p1mrfhop', 'vekibwgj', 'w08e1crn', 'ztqee46x']
    >>> data_source = asyncio.run(NeurobankSource.create(url, stimuli, responses))
    >>> builder = DatasetBuilder()
    >>> builder.set_data_source(data_source)
    >>> builder.load_responses()
    >>> builder.bin_responses(time_step=0.005) # 5 ms
    >>> builder.add_stimuli(Gammatone(
    ...     window_time=0.005,
    ...     frequency_bin_count=50,
    ...     min_frequency=500,
    ...     max_frequency=8000,
    ...     log_transform_compress=1,
    ... ))
    >>> builder.create_time_lags(tau=0.3)

    -->
    >>> dataset = builder.get_dataset()
    >>> X, Y = dataset[20:30]
    >>> X.shape
    (4600, 60)
    >>> Y.shape
    (4600, 50)

    You can also manually select directly from the DataFrames.
    """

    responses: Optional[pd.DataFrame] = None
    """Columns correspond to pprox files
    """
    stimuli: Optional[pd.DataFrame] = None
    """"""
    trial_data: Optional[pd.DataFrame] = None
    """Each row corresponds to a trial and each column corresponds to a key in the
    pprox
    """
    time_step: Optional[float] = None
    """granularity of time"""
    stimuli_format: Optional[StimuliFormat] = None
    """format of the stimuli"""

    def _get_stimuli(self) -> pd.DataFrame:
        if self.stimuli is None:
            raise InvalidConstructionSequence("add_stimuli")
        return self.stimuli

    def _get_time_step(self) -> float:
        if self.time_step is None:
            raise InvalidConstructionSequence("bin_responses")
        return self.time_step

    def _set_time_step(self, time_step: float):
        if self.time_step is None:
            self.time_step = time_step
        else:
            raise TimestepSetTwice()

    def _get_trial_data(self) -> pd.DataFrame:
        if self.trial_data is None:
            raise InvalidConstructionSequence("load_responses")
        return self.trial_data

    def _get_responses(self) -> pd.DataFrame:
        if self.responses is None:
            raise InvalidConstructionSequence("load_responses")
        return self.responses

    def __getitem__(self, key):
        """
        get numpy arrays representing the responses and the stimuli
        at the given pandas index range. The array dimensions are (time, neuron * lag).
        The dimensions of the stimulus will depend on the StimuliFormat chosen.
        """
        selection = (
            self._get_responses()
            .index.get_level_values(0)
            .unique()
            .to_series()
            .loc[key]
        )
        responses = self._get_responses().loc[selection].to_numpy()
        try:
            stimuli_index = self._get_trial_data().loc[selection]["stimulus.name"]
        except KeyError:
            stimuli_index = key
        stimuli = self._get_stimuli_format().to_values(
            self._get_stimuli().loc[stimuli_index]
        )
        return responses, stimuli

    def _get_stimuli_format(self) -> StimuliFormat:
        if self.stimuli_format is None:
            raise InvalidConstructionSequence("add_stimuli")
        return self.stimuli_format

    def to_steps(self, time_in_seconds):
        """Converts a time in seconds to a time in steps"""
        return int(time_in_seconds / self._get_time_step())


class DatasetBuilder:
    """Construct instances of the `Dataset` class using the [builder
    pattern](https://refactoring.guru/design-patterns/builder)
    """

    data_source: DataSource
    tau: Optional[float]
    basis: Optional[np.ndarray]

    def __init__(self):
        self._dataset = Dataset()
        self.data_source = _EmptySource()

    def set_data_source(self, data_source: DataSource):
        self.data_source = data_source

    def load_responses(self, ignore_columns: Optional[List[str]] = None):
        if ignore_columns is None:
            ignore_columns = []
        clusters = self.data_source.get_responses()
        assert len(clusters) > 0, "no clusters"
        trial_data = pd.concat(
            {
                k: pd.json_normalize(v["pprox"])
                .rename(columns={"index": "trial"})
                .set_index("trial")
                for k, v in clusters.items()
            },
            axis=1,
        )
        trial_data.columns = trial_data.columns.reorder_levels(order=[1, 0])
        self._dataset.responses = trial_data["events"].rename_axis(columns="neuron")
        ignore_columns.append("events")
        trial_data = trial_data.drop(columns=ignore_columns, level=0)
        single_trial = self._aggregate_trials(trial_data)
        assert single_trial is not None
        _, trial_data = single_trial
        self._dataset.trial_data = trial_data

    @staticmethod
    def _aggregate_trials(
        trial_data: pd.DataFrame,
    ) -> Optional[Tuple[str, pd.DataFrame]]:
        """combine corresponding trials across different pprox files

        if all trials contain the same data, return one trial
        otherwise, raise a IncompatibleTrialError
        """
        trials = trial_data.groupby(axis=1, level=1)
        first = None
        for name, t in trials:
            t = t.droplevel(axis=1, level=1)
            if first is None:
                first = (name, t)
            first_name, first_df = first
            if not first_df.equals(t):
                raise IncompatibleTrialError({first_name: first_df, name: t})
        return first

    def bin_responses(self, time_step: float = 0.005, normalize: bool = True):
        """
        transform a point process into bins of size `time_step` containing
        the number of events (or if `normalize` is True, the rate of events)
        that occurred within the time bin.
        """
        self._dataset._set_time_step(time_step)
        spike_times = self._dataset._get_responses()

        def cut(df):
            start, stop = df.name
            # np.arange does not include the end of the interval,
            # so we make the end one time_step later
            edges = np.arange(start, stop + time_step, time_step)
            arr = np.block(
                df[spike_times.columns]
                .applymap(
                    lambda spikes: np.histogram(spikes, bins=edges)[0].reshape(-1, 1)
                )
                .to_numpy()
                .tolist()
            )
            if normalize:
                arr = arr / time_step
            return pd.DataFrame(
                arr,
                index=pd.MultiIndex.from_product(
                    [df.index, edges[:-1]], names=["trial", "time"]
                ),
                columns=spike_times.columns,
            )

        self._dataset.responses = (
            spike_times.join(self._dataset._get_trial_data()[["interval"]])
            .groupby("interval", group_keys=False)
            .apply(cut)
        ).sort_index()

    def add_stimuli(
        self, stimuli_format: StimuliFormat, time_step: Optional[float] = None
    ):
        """
        Add a dataframe containing formatted stimuli for each
        stimulus associated with a trial

        Consult documentation for `preconstruct.stimuliformats` for details.

        Specify `time_step` only if not already specified in `bin_responses`, if
        you are only building the stimuli DataFrame, for example.

        ###### example
        <!--
        >>> import asyncio
        >>> from preconstruct.sources import NeurobankSource
        >>> responses = ['P120_1_1_c92']
        >>> url = 'https://gracula.psyc.virginia.edu/neurobank/'
        >>> stimuli = ['c95zqjxq', 'g29wxi4q', 'igmi8fxa', 'jkexyrd5', 'l1a3ltpy',
        ...         'mrel2o09', 'p1mrfhop', 'vekibwgj', 'w08e1crn', 'ztqee46x']
        >>> data_source = asyncio.run(NeurobankSource.create(url, stimuli, responses))
        >>> builder = DatasetBuilder()
        >>> builder.set_data_source(data_source)
        >>> builder.load_responses()
        >>> builder.bin_responses(time_step=0.005) # 5 ms

        -->
        >>> from preconstruct.stimuliformats import Gammatone
        >>> builder.add_stimuli(Gammatone(
        ...     window_time=0.001,
        ...     frequency_bin_count=50,
        ...     min_frequency=500,
        ...     max_frequency=8000,
        ... ))
        """
        if time_step is not None:
            self._dataset._set_time_step(time_step)
        self._dataset.stimuli_format = stimuli_format
        self._dataset.stimuli = stimuli_format.create_dataframe(
            self.data_source,
            self._dataset._get_time_step(),
            self._dataset._get_trial_data()[["stimulus.name", "stimulus.interval"]]
            .drop_duplicates()
            .set_index("stimulus.name")["stimulus.interval"],
        )

    def create_time_lags(self, tau: float = 0.300, basis: Optional[Basis] = None):
        """
        `tau`: length of window (in secs) to consider in prediction
        `basis`: an instance of a class that inherits from
        `preconstruct.basisfunctions.Basis`, initialized with the dimension
        of the projection

        ###### example
        <!--
        >>> import asyncio
        >>> from preconstruct.sources import NeurobankSource
        >>> from preconstruct.stimuliformats import Gammatone
        >>> responses = ['P120_1_1_c92']
        >>> url = 'https://gracula.psyc.virginia.edu/neurobank/'
        >>> stimuli = ['c95zqjxq', 'g29wxi4q', 'igmi8fxa', 'jkexyrd5', 'l1a3ltpy',
        ...         'mrel2o09', 'p1mrfhop', 'vekibwgj', 'w08e1crn', 'ztqee46x']
        >>> data_source = asyncio.run(NeurobankSource.create(url, stimuli, responses))
        >>> builder = DatasetBuilder()
        >>> builder.set_data_source(data_source)
        >>> builder.load_responses()
        >>> builder.bin_responses(time_step=0.005) # 5 ms
        >>> builder.add_stimuli(Gammatone())


        -->
        >>> from preconstruct.basisfunctions import RaisedCosineBasis
        >>> builder.create_time_lags(tau=0.3, basis=RaisedCosineBasis(30))


        """
        self.tau = tau
        self.basis = None
        if basis is not None:
            window_length = self._dataset.to_steps(self.tau)
            self.basis = basis.get_basis(window_length)
        # figure out if trials have been pooled and adjust accordingly
        common_index = self._dataset._get_responses().index.names[0]
        self._dataset.responses = (
            self._dataset._get_responses()
            .groupby(level=0, axis=1)
            .apply(
                lambda neuron: self._dataset._get_trial_data()[
                    ["stimulus.interval", "stimulus.name"]
                ]
                .merge(
                    neuron[neuron.columns[0]]  # convert to series
                    .rename("events")
                    .reset_index(level="time"),
                    on=common_index,
                    how="left",
                )
                .join(
                    self._dataset._get_stimuli()
                    .groupby(level=0)
                    .apply(lambda x: x.shape[0])
                    .rename("stimulus.length"),
                    on="stimulus.name",
                )
                .groupby(common_index)
                .apply(self._stagger)
            )
        )

    def _stagger(self, df):
        df = df.reset_index()
        [stimulus_interval] = df["stimulus.interval"].unique()
        stim_start, _ = stimulus_interval
        start = df["time"][df["time"] >= stim_start].idxmin()
        window_length = self._dataset.to_steps(self.tau)
        [stimulus_length] = df["stimulus.length"].unique()
        stop = start + stimulus_length
        events = df["events"]
        assert len(events) >= stop - 1 + window_length
        time_lagged = hankel(
            events[start:stop], events[stop - 1 : stop - 1 + window_length]
        )
        columns = pd.Series(
            np.arange(window_length) * self._dataset._get_time_step(), name="offset"
        )
        if self.basis is not None:
            time_lagged = np.dot(time_lagged, self.basis)
            columns = None
        return pd.DataFrame(
            time_lagged,
            index=df["time"].iloc[start:stop],
            columns=columns,
        )

    def pool_trials(self):
        """Pool spikes across trials"""
        neurons = self._dataset._get_responses().columns
        if (
            not self._dataset._get_trial_data()
            .groupby("stimulus.name")["stimulus.interval"]
            .apply(lambda ser: len(ser.unique()) == 1)
            .all()
        ):
            raise InconsistentStimulusInterval

        responses = self._dataset._get_responses().reset_index()
        responses["stimulus.name"] = responses["trial"].map(
            self._dataset._get_trial_data()["stimulus.name"]
        )
        del responses["trial"]
        responses = responses.set_index(["stimulus.name", "time"]).sort_index()
        self._dataset.responses = responses.groupby(["stimulus.name", "time"]).agg(
            "sum"
        )

    def get_dataset(self) -> Dataset:
        """Return the fully constructed `Dataset` object"""
        return self._dataset


class _EmptySource(DataSource):
    def _get_raw_responses(self):
        self._raise()

    def get_stimuli(self):
        self._raise()

    @staticmethod
    def _raise():
        raise InvalidConstructionSequence("DatasetBuilder.set_data_source")


class InvalidConstructionSequence(Exception):
    """Indicates that the methods of a DatasetBuilder have been called in an invalid order"""

    def __init__(self, required_method):
        super().__init__()
        self.required_method = required_method

    def __str__(self) -> str:
        return (
            f"invalid construction sequence: must call `{self.required_method}` first"
        )


class IncompatibleTrialError(Exception):
    """Raised when corresponding trials from different pprox files differ

    If the differing trial data is expected, you can choose to not include
    specific keys using the `ignore_columns` argument in
    `DatasetBuilder.load_responses`.
    """

    def __init__(self, trial_pair: Dict[str, pd.DataFrame]):
        super().__init__()
        self.trial_pair = trial_pair

    def __str__(self) -> str:
        a, b = tuple(self.trial_pair.keys())
        return (
            f"at least two trials contained conflicting data: {a}, {b}\n"
            f"{self.trial_pair[a].compare(self.trial_pair[b])}"
        )


class InconsistentStimulusInterval(Exception):
    """Raised when stimulus.interval differs between trials that are to be combined"""

    def __str__(self) -> str:
        return (
            "if you want to pool trials, every stimulus presentation must have"
            " the same stimulus.interval"
        )


class IncompatibleStimuliFormat(Exception):
    def __init__(self, format):
        self.format = format

    def __str__(self) -> str:
        return (
            f"the StimuliFormat you have picked ({self.format})"
            " is not a subclass of `SameTimeIndexAsResponse`"
            " so you can't call this function"
        )


class TimestepSetTwice(Exception):
    def __init__(self):
        super().__init__(self, "You can only set time_step once")



"""Basis functions for projecting binned spikes.

"""
from abc import ABC, abstractmethod
import numpy as np


class Basis(ABC):
    @abstractmethod
    def __init__(self, num_basis_functions: int):
        pass

    @abstractmethod
    def get_basis(self, num_timesteps: int) -> np.ndarray:
        """Returns a ndarray with shape
        (num_timesteps, num_basis_functions)

        NB: The assumption is that this matrix will be applied to hankel
        matrices, with lag decreasing across columns.

        basis vectors must have short lags at the top,
        matching the way the Hankel matrix is constructed
        """


class RaisedCosineBasis(Basis):
    """Make a nonlinearly stretched basis consisting of raised cosines.
    """

    _MIN_OFFSET = 1e-20

    def __init__(self, num_basis_functions, linearity_factor=10):
        """
        linearity_factor: offset for nonlinear stretching of x axis
                          (larger values means closer to linear spacing)
        """
        self.num_basis_functions = num_basis_functions
        self.linearity_factor = linearity_factor

    @staticmethod
    def raised_cos(center, domain, peak_spacing) -> np.ndarray:
        """Evaluates a raised cosine (which looks similar to a Gaussian
        curve) over a discrete domain given by `domain`, centered at
        `center` and with spread related to `peak_spacing`.
        """
        cos_input = np.clip((domain - center) * np.pi / peak_spacing / 2, -np.pi, np.pi)
        return (np.cos(cos_input) + 1) / 2

    def get_basis(self, num_timesteps) -> np.ndarray:
        # nonlinearity for stretching x axis
        nonlinearity = lambda x: np.log(x + self.linearity_factor + self._MIN_OFFSET)

        first_peak = nonlinearity(0)
        last_peak = nonlinearity(num_timesteps * (1 - 1.5 / self.num_basis_functions))
        peak_centers = np.linspace(first_peak, last_peak, self.num_basis_functions)
        log_domain = nonlinearity(np.arange(num_timesteps))
        peak_spacing = (last_peak - first_peak) / (self.num_basis_functions - 1)
        basis = np.column_stack(
            [self.raised_cos(c, log_domain, peak_spacing) for c in peak_centers]
        )
        basis /= np.linalg.norm(basis, axis=0)
        return basis



"""Data sources

## Examples

Let's build a simple application that accepts input in the form of a DataSource.

>>> import numpy as np
>>> from preconstruct.sources import DataSource, MemorySource
>>> def stim_resp_summary(data_source: DataSource):
...     stimuli = data_source.get_stimuli()
...     print('stimuli sample rates:')
...     for name, (fs, _samples) in stimuli.items():
...         print(f'{name}: {fs}')
...     responses = data_source.get_responses()
...     print('responses trial count:')
...     for name, pprox in responses.items():
...         print(f'{name}: {len(pprox["pprox"])}')

>>> responses = ['P120_1_1_c92']
>>> url = 'https://gracula.psyc.virginia.edu/neurobank/'
>>> stimuli = ['c95zqjxq', 'g29wxi4q', 'igmi8fxa', 'jkexyrd5', 'l1a3ltpy',
...         'mrel2o09', 'p1mrfhop', 'vekibwgj', 'w08e1crn', 'ztqee46x']
>>> data_source = asyncio.run(NeurobankSource.create(url, stimuli, responses))
>>> stim_resp_summary(data_source)
stimuli sample rates:
c95zqjxq: 44100
g29wxi4q: 44100
igmi8fxa: 44100
jkexyrd5: 44100
l1a3ltpy: 44100
mrel2o09: 44100
p1mrfhop: 44100
vekibwgj: 44100
w08e1crn: 44100
ztqee46x: 44100
responses trial count:
P120_1_1_c92: 100
"""
import asyncio
from collections import defaultdict
from typing import Dict, List, Tuple, Optional, Generator, Union, TypedDict, Set
from abc import ABC, abstractmethod
from pathlib import Path
from glob import glob
import json
from urllib.parse import urljoin, urlparse

import numpy as np
import aiohttp
import parse
from scipy.io import wavfile
from appdirs import user_cache_dir


from preconstruct import APP_NAME, APP_AUTHOR


class StimConfig(TypedDict):
    """type annotation for information about a stimulus presentation"""

    name: str
    interval: Tuple[float]


class Trial(TypedDict):
    """type annotation for a trial"""

    events: List[float]
    interval: Tuple[float]
    stimulus: StimConfig
    offset: Optional[float]
    index: Optional[int]


class Pprox(TypedDict):
    """type annotation for a collection of trials"""

    pprox: List[Trial]


Wav = Tuple[int, np.ndarray]


class DataSource(ABC):
    """Abstract class for data sources to inherit from

    The methods on this class will be available from all the child classes
    in the module, but you cannot directly instantiate this class.

    Concrete child implementations must define how to load data (e.g. local path, URL).
    By inheriting from this class, they provide a standardized interface to
    access the data
    """

    @abstractmethod
    def _get_raw_responses(self) -> Dict[str, Pprox]:
        """returns dictionary mapping names of files to pprox data

        pprox may be in the incorrect format
        """

    def get_responses(self) -> Dict[str, Pprox]:
        """returns dictionary mapping names of files to stimtrial pprox data"""
        responses = self._get_raw_responses()
        stimuli = self.get_stimuli()
        # remove trials for stimuli that are not in the list
        durations = {name: len(s) / fs for name, (fs, s) in stimuli.items()}
        responses = _fix_pprox(responses, durations)
        for nrn_name, nrn in responses.items():
            nrn["pprox"] = [trial for trial in nrn["pprox"] if trial["stimulus"]["name"] in stimuli]
        return responses

    @abstractmethod
    def get_stimuli(self) -> Dict[str, Wav]:
        """returns dictionary mapping names of files to (sample_rate, samples) tuples"""

    def __eq__(self, other) -> bool:
        return (self.get_responses() == other.get_responses()) and (
            self.get_stimuli() == other.get_stimuli()
        )

    def stimuli_names_from_pprox(self) -> Set[str]:
        """returns a set of all stimuli used in the responses"""
        return set(self._stimuli_generator())

    def _stimuli_generator(self) -> Generator[str, None, None]:
        durations = defaultdict(lambda: 0.0)
        responses = _fix_pprox(self._get_raw_responses(), durations)
        for pprox in responses.values():
            for trial in pprox["pprox"]:
                yield trial["stimulus"]["name"]


IdentifierCollection = Union[str, Path, List[str]]
"""
a list of resource IDs, or path to a file containing such a list
"""


class FsSource(DataSource):
    """Loads data from local File System (FS)"""

    def __init__(
        self,
        responses: Union[str, Path],
        stimuli: Union[str, Path],
        stimuli_names: Optional[IdentifierCollection] = None,
        cluster_list: Optional[IdentifierCollection] = None,
    ):
        """initialize with path to data

        Paths are specified with a format string, where `{}` is used to
        represent the part of the path/filename that determines the identity of
        the file contents. This way, files identifiers do not need to contain
        redundant information, such as file extension. If a list of identifers
        is not specified, the wildcard `*` will be substituted for `{}` to load
        all matching files.

        `responses`: e.g. `'pprox/P120_1_1_{}.pprox'`

        `stimuli`: e.g. `'wav/{}.wav'`

        `cluster_list`: optional `IdentifierCollection`, defaults to all files present

        `stimuli_names`: optional `IdentifierCollection`, defaults to all files present
        """
        if not isinstance(responses, str):
            responses = str(responses)
        if not isinstance(stimuli, str):
            stimuli = str(stimuli)
        self.responses = responses
        self.cluster_list = self._into_list(cluster_list)
        self.stimuli = stimuli
        self.stimuli_names = self._into_list(stimuli_names)

    @staticmethod
    def _into_list(resource_ids: Optional[IdentifierCollection]) -> Optional[List[str]]:
        if resource_ids is None:
            return None
        if isinstance(resource_ids, list):
            return resource_ids
        if isinstance(resource_ids, (str, Path)):
            with open(resource_ids, "r") as fd:
                return fd.read().splitlines()
        raise ValueError("input should be an IdentifierCollection")

    def _get_raw_responses(self):
        return self._load_pprox(
            self.responses,
            cluster_names=self.cluster_list,
        )

    def get_stimuli(self) -> Dict[str, Wav]:
        return self._load_stimuli(self.stimuli, self.stimuli_names)

    @staticmethod
    def _load_stimuli(path_format, stimuli_names=None):
        """Load wav files into a dictionary

        stimuli_names: iterable of strings
        path_format: path containing "{}" which will be replaced by
                     each of `stimuli_names`

        returns: dictionary mapping stimuli_names to (sample_rate, samples)
        """
        stimuli = {}
        for name, filename in FsSource._get_filenames(path_format, stimuli_names):
            sample_rate, samples = wavfile.read(filename)
            stimuli[name] = (sample_rate, samples)
        return stimuli

    @staticmethod
    def _load_pprox(path_format, cluster_names=None):
        """Load pprox files into a dictionary
        path_format: path containing "{}" which will be globbed
                     for all matching files and then parsed as
                     JSON
        cluster_names (iterable): if specified, load only the clusters
                       with the given names
        returns: dictionary mapping the part of the filenames
                 represented by "{}" to their parsed contents
        """
        clusters = {}
        for name, path in FsSource._get_filenames(path_format, cluster_names):
            with open(path, "r") as pprox_file:
                try:
                    json_data = json.load(pprox_file)
                except UnicodeDecodeError as exc:
                    raise ValueError(
                        "could not load pprox files as text data"
                        " (are you putting stimuli where responses should go?)"
                    ) from exc
                clusters[name] = json_data
        return clusters

    @staticmethod
    def _get_filenames(path_format, names):
        assert path_format.find("{}") != -1, (
            "paths should include empty braces where a wildcard"
            " would go.\n"
            'For example, "{}.pprox" will load all files in the current'
            ' directory that end in ".pprox". The contents of each file'
            ' will be named with the part of the filename before ".pprox".'
        )
        parser = parse.compile(path_format)
        if names is None:
            filenames = glob(path_format.format("*"))
            names = map(lambda p: parser.parse(p)[0], filenames)
        else:
            filenames = map(path_format.format, names)
        return zip(names, filenames)


class NeurobankSource(FsSource):
    """Downloads data from Neurobank and caches the files

    `NeurobankSource` uses asynchronous network calls in order to download
    resources in parallel. As a result, it must be initialized in a special
    way:
    ```
    >>> import asyncio
    >>> from preconstruct.sources import NeurobankSource
    >>> stimuli = ['ztqee46x', '00oagdl5', 'g29wxi4q', 'mrel2o09', 'vekibwgj', \
        'l1a3ltpy', 'igmi8fxa', 'c95zqjxq', 'w08e1crn', 'jkexyrd5', 'p1mrfhop', ]
    >>> responses = ['P120_1_1_c92']
    >>> url = 'https://gracula.psyc.virginia.edu/neurobank/'
    >>> data_source = asyncio.run(NeurobankSource.create(url, stimuli, responses))

    ```

    In a Jupyter Notebook, you can skip importing `asyncio` and just run:
    ```
    data_source = await NeurobankSource.create(url, stimuli, responses)
    ```
    """

    _DOWNLOAD_FORMAT = "resources/{}/download"

    @classmethod
    async def create(
        cls,
        neurobank_registry: str,
        wav_ids: IdentifierCollection,
        pprox_ids: IdentifierCollection,
        infer_stimuli=False,
    ):
        """
        `neurobank_registry`: URL of Neurobank instance
        `infer_stimuli`: whether to use stimuli names inferred from the responses
        """
        url_format = urljoin(neurobank_registry, cls._DOWNLOAD_FORMAT)
        pprox_ids_ = cls._into_list(pprox_ids)
        wav_ids_ = cls._into_list(wav_ids)
        parsed_url = urlparse(neurobank_registry)
        cache_dir = Path(user_cache_dir(APP_NAME, APP_AUTHOR)) / parsed_url.netloc
        self = NeurobankSource(url_format, pprox_ids_, wav_ids_, cache_dir)
        responses_dir = self.cache_dir / "responses"
        stimuli_dir = self.cache_dir / "stimuli"
        responses_dir.mkdir(parents=True, exist_ok=True)
        stimuli_dir.mkdir(parents=True, exist_ok=True)
        await self._download_all()
        super().__init__(
            self,
            responses_dir / "{}",
            stimuli_dir / "{}",
            cluster_list=self.pprox_ids,
            stimuli_names=self.wav_ids,
        )
        if infer_stimuli:
            self.wav_ids = list(self.stimuli_names_from_pprox())
            self.stimuli_names = self.wav_ids
            await self._download_stimuli()
        return self

    def __init__(self, url_format, pprox_ids, wav_ids, cache_dir):
        """`NeurobankSource` cannot be directly initialized.
        Use `NeurobankSource.create` instead."""
        self.url_format = url_format
        self.pprox_ids = pprox_ids
        self.wav_ids = wav_ids
        self.cache_dir = cache_dir

    async def _download(self, resource_id, session, folder):
        url = self.url_format.format(resource_id)
        path = Path(self.cache_dir / folder / resource_id)
        if path.is_file():
            return
        async with session.get(str(url)) as resp:
            with open(path, "wb") as fd:
                async for chunk in resp.content.iter_chunked(1024):
                    fd.write(chunk)

    async def _download_all(self):
        await asyncio.gather(self._download_responses(), self._download_stimuli())

    async def _download_responses(self):
        async with aiohttp.ClientSession(raise_for_status=True) as session:
            await asyncio.gather(
                *[self._download(url, session, "responses") for url in self.pprox_ids],
            )

    async def _download_stimuli(self):
        async with aiohttp.ClientSession(raise_for_status=True) as session:
            await asyncio.gather(
                *[self._download(url, session, "stimuli") for url in self.wav_ids],
            )


class MemorySource(DataSource):
    """Loads data from given dictionaries"""

    def __init__(self, responses: dict, stimuli):
        self.responses = responses
        self.stimuli = stimuli

    def _get_raw_responses(self):
        return self.responses

    def get_stimuli(self):
        return self.stimuli


def _fix_pprox(responses: Dict[str, Pprox], durations: Dict[str, float]):
    for name, json_data in responses.items():
        if json_data.get("$schema") == "https://meliza.org/spec:2/stimtrial.json#":
            pass
        # if the pprox does not conform to the stimtrial spec,
        # we need to try to guess what format it is, and modify it
        # accordingly
        elif json_data.get("experiment") == "induction":
            _ar_data_shim(json_data, durations)
            _cn_data_shim(json_data, durations)
        elif json_data.get("protocol") in ["songs", "chorus"]:
            _cn_data_shim(json_data, durations)
        else:
            raise UnrecognizedPproxFormat(name)
        for i, trial in enumerate(json_data["pprox"]):
            trial["index"] = i
            trial["interval"] = tuple(trial["interval"])
            trial["stimulus"]["interval"] = tuple(trial["stimulus"]["interval"])
    return responses


class UnrecognizedPproxFormat(Exception):
    """Exception raised when the needed metadata could not be found in a pprox file"""

    def __init__(self, name):
        super().__init__()
        self.name = name

    def __str__(self):
        return (
            f"could not detect pprox format for file {self.name}, try making sure"
            " the pprox conforms to stimtrial and that '$schema' is set"
        )


def _ar_data_shim(json_data, durations):
    """reformat data from auditory-restoration project format to colony-noise project format"""
    for trial in json_data["pprox"]:
        _rename_key(trial, "event", "events")
        if trial.get("units") == "ms":
            del trial["units"]
            trial["events"] = [x / 1000 for x in trial["events"]]
        trial["stim"] = f"{trial['stimulus']}_{trial['condition']}"
        _rename_key(trial, "trial", "index")
        _rename_key(trial, "stimulus", "base_stimulus")
        # we want to put the `interval` information in the cn_pprox format
        # so we list absolute recording start and stop with a sample rate of 1
        trial["recording"] = {
            "start": -0.5,
            "stop": trial["stim_on"] + durations[trial["stim"]] + 0.5,
        }
    json_data["entry_metadata"] = {"sampling_rate": 1}
    del json_data["experiment"]
    json_data["protocol"] = "songs"
    return json_data


def _cn_data_shim(json_data, durations):
    """reformat data from colony-noise project format to stimtrial format"""
    metadata = json_data["entry_metadata"]
    if isinstance(metadata, list):
        metadata = metadata[0]
    sampling_rate = metadata["sampling_rate"]
    del json_data["entry_metadata"]
    json_data["$schema"] = "https://meliza.org/spec:2/stimtrial.json#"
    del json_data["protocol"]
    for trial in json_data["pprox"]:
        offset = trial.get("offset") or 0
        trial["interval"] = (
            (trial["recording"]["start"] / sampling_rate) - offset,
            (trial["recording"]["stop"] / sampling_rate) - offset,
        )
        del trial["recording"]
        stim_off = durations[trial["stim"]]
        trial["stimulus"] = {
            "name": trial["stim"],
            "interval": (trial["stim_on"], trial["stim_on"] + stim_off),
        }
        del trial["stim"]
        del trial["stim_on"]
    return json_data


def _rename_key(obj, old_name, new_name):
    obj[new_name] = obj[old_name]
    del obj[old_name]



import pytest
from preconstruct.sources import NeurobankSource, MemorySource


@pytest.fixture
def nbank_stimuli():
    return [
        "ztqee46x",
        "g29wxi4q",
        "mrel2o09",
        "vekibwgj",
        "l1a3ltpy",
        "igmi8fxa",
        "c95zqjxq",
        "w08e1crn",
        "jkexyrd5",
        "p1mrfhop",
    ]


@pytest.fixture
def nbank_responses():
    return ["P120_1_1_c92"]


@pytest.fixture
def nbank_url():
    return "https://gracula.psyc.virginia.edu/neurobank/"


async def test_neurobank(nbank_responses, nbank_stimuli, nbank_url):
    source = await NeurobankSource.create(nbank_url, nbank_stimuli, nbank_responses)
    assert len(source.get_responses()) == 1
    assert len(source.get_stimuli()) == len(nbank_stimuli)


async def test_neurobank_autonames(nbank_responses, nbank_stimuli, nbank_url):
    source = await NeurobankSource.create(
        nbank_url, [], nbank_responses, infer_stimuli=True
    )
    assert len(source.get_responses()) == 1
    assert len(source.get_stimuli()) == len(nbank_stimuli)


async def test_neurobank_from_file(nbank_responses, nbank_stimuli, nbank_url, tmp_path):
    stim_file = tmp_path / "stims"
    stim_file.write_text("\n".join(nbank_stimuli))
    resp_file = tmp_path / "resps"
    resp_file.write_text("\n".join(nbank_responses))
    source = await NeurobankSource.create(nbank_url, stim_file, resp_file)
    assert len(source.get_responses()) == 1
    assert len(source.get_stimuli()) == len(nbank_stimuli)


@pytest.mark.skip(
    "bad thing to test for. they will never share the identical set of keys"
)
def test_all_formats_equiv(stimuli, cn_pprox, ar_pprox, stimtrial_pprox):
    cn_data = MemorySource(cn_pprox, stimuli)
    ar_data = MemorySource(ar_pprox, stimuli)
    stimtrial = MemorySource(stimtrial_pprox, stimuli)
    print(stimtrial.get_responses())
    print(ar_data.get_responses())
    assert stimtrial == cn_data
    assert stimtrial == ar_data


async def test_show_stimuli(nbank_responses, nbank_stimuli, nbank_url):
    source = await NeurobankSource.create(nbank_url, [], nbank_responses)
    assert source.stimuli_names_from_pprox() == set(nbank_stimuli)


async def test_bad_order(nbank_responses, nbank_stimuli, nbank_url):
    source = await NeurobankSource.create(nbank_url, nbank_responses, nbank_stimuli)
    with pytest.raises(ValueError):
        source.get_responses()



import pytest
import numpy as np
import pandas as pd

from preconstruct import DatasetBuilder
from preconstruct.sources import NeurobankSource, MemorySource
from preconstruct.stimuliformats import Gammatone
from preconstruct.dataset import *


@pytest.fixture
def mem_data_source(stimtrial_pprox, stimuli):
    responses = stimtrial_pprox
    return MemorySource(responses, stimuli)


@pytest.fixture
async def real_data_source():
    responses = ["P120_1_1_c92", "P120_1_1_c89"]
    stimuli = [
        "c95zqjxq",
        "g29wxi4q",
        "igmi8fxa",
        "jkexyrd5",
        "l1a3ltpy",
        "mrel2o09",
        "p1mrfhop",
        "vekibwgj",
        "w08e1crn",
        "ztqee46x",
    ]
    url = "https://gracula.psyc.virginia.edu/neurobank/"
    return await NeurobankSource.create(url, stimuli, responses)


def test_building(mem_data_source):
    builder = DatasetBuilder()
    builder.set_data_source(mem_data_source)
    builder.load_responses()

    time_step = 0.001
    builder.bin_responses(time_step=time_step, normalize=False)
    neuron = "neuron_1"
    trial_index = 0
    trial = mem_data_source.get_responses()[neuron]["pprox"][trial_index]
    stimulus = trial["stimulus"]
    recording_duration = (trial["interval"][1] - trial["interval"][0]) / time_step
    binned = np.zeros(int(recording_duration))
    binned[200] = binned[1000] = binned[1500] = 1
    actual_binned = builder._dataset._get_responses()[neuron].loc[trial_index]
    assert actual_binned.shape == binned.shape
    assert np.array_equiv(binned, actual_binned.to_numpy())

    builder.add_stimuli(Gammatone())
    spectrogram = builder._dataset._get_stimuli().loc[stimulus["name"]]
    spectrogram_length = spectrogram.shape[0]

    tau = 0.3
    builder.create_time_lags(tau=tau)
    actual_lagged = builder._dataset._get_responses()[neuron].loc[trial_index]
    print(spectrogram)
    shape = (spectrogram_length, int(tau / time_step))
    assert actual_lagged.shape == shape
    dataset = builder.get_dataset()
    X, Y = dataset[[0]]
    assert len(X) == len(Y)


def test_pool_trials(real_data_source):
    builder = DatasetBuilder()
    builder.set_data_source(real_data_source)
    builder.load_responses()
    builder.bin_responses()
    builder.add_stimuli(Gammatone())
    builder.create_time_lags()
    neurons = builder._dataset._get_responses().columns
    builder.pool_trials()
    assert (builder._dataset._get_responses().columns == neurons).all()
    dataset = builder.get_dataset()
    X, Y = dataset[["ztqee46x"]]
    assert len(X) == len(Y)
    assert isinstance(dataset._get_responses().columns, pd.MultiIndex)


def test_pool_trials_before_lag(real_data_source):
    builder = DatasetBuilder()
    builder.set_data_source(real_data_source)
    builder.load_responses()
    builder.bin_responses()
    builder.add_stimuli(Gammatone())
    builder.pool_trials()
    builder.create_time_lags()
    dataset_pool_first = builder.get_dataset()
    # compare with pooling after lag
    builder = DatasetBuilder()
    builder.set_data_source(real_data_source)
    builder.load_responses()
    builder.bin_responses()
    builder.add_stimuli(Gammatone())
    builder.create_time_lags()
    builder.pool_trials()
    dataset_pool_second = builder.get_dataset()
    assert np.all(
        [
            np.array_equal(a, b)
            for a, b in zip(dataset_pool_first[:], dataset_pool_second[:])
        ]
    )


async def test_margot_data():
    from preconstruct import sources, dataset, basisfunctions
    from sklearn.linear_model import RidgeCV, Ridge

    responses = ["P4_p1r2_ch20_c31", "O129_p1r2_ch19_c3", "P4_p1r2_ch22_c23"]
    stimuli = []
    url = "https://gracula.psyc.virginia.edu/neurobank/"
    test_source = await (sources.NeurobankSource.create(url, stimuli, responses))
    stimuli = list(test_source.stimuli_names_from_pprox())
    data_source = await (sources.NeurobankSource.create(url, stimuli, responses))
    builder = dataset.DatasetBuilder()
    builder.set_data_source(data_source)
    print("loading responses")
    builder.load_responses(ignore_columns=["category"])
    print("binning responses")
    builder.bin_responses(time_step=0.001)  # 5 ms
    print("computing spectrograms")
    builder.add_stimuli(
        Gammatone(
            window_time=0.0025,
            frequency_bin_count=30,
            min_frequency=500,
            max_frequency=8000,
            log_transform_compress=1,
        )
    )
    basis = basisfunctions.RaisedCosineBasis(30, linearity_factor=30)
    print("creating design matrices")
    builder.pool_trials()
    builder.create_time_lags(tau=0.3, basis=basis)
    dataset = builder.get_dataset()
    training_stimuli = stimuli
    print(builder._dataset._get_responses())
    X, Y = dataset[training_stimuli]
    print("X.shape:", X.shape)
    print("Y.shape:", Y.shape)
    print(
        dataset._get_responses().loc[training_stimuli[0]].index
        - dataset._get_stimuli().loc[training_stimuli[0]].index
    )
    assert np.allclose(
        dataset._get_responses().loc[training_stimuli[0]].index,
        dataset._get_stimuli().loc[training_stimuli[0]].index,
        atol=1e-2,
    )
    estimator = Ridge(alpha=8.59)
    estimator.fit(X, Y)
    print("model score:", estimator.score(X, Y))


def test_stimuli_only(real_data_source):
    builder = DatasetBuilder()
    builder.set_data_source(real_data_source)
    builder.load_responses()
    frequency_bin_count = 50
    with pytest.raises(InvalidConstructionSequence, match="bin_responses"):
        builder.add_stimuli(Gammatone(frequency_bin_count=frequency_bin_count))
    builder.add_stimuli(
        Gammatone(frequency_bin_count=frequency_bin_count), time_step=0.05
    )
    with pytest.raises(TimestepSetTwice):
        builder.bin_responses(0.05)
    dataset = builder.get_dataset()
    Y = dataset._get_stimuli()
    assert Y.shape[1] == frequency_bin_count



import pytest
from preconstruct import DatasetBuilder
from preconstruct.sources import MemorySource
from preconstruct.stimuliformats import *


def test_spectrogram(real_data_source):
    builder = DatasetBuilder()
    builder.set_data_source(real_data_source)
    builder.load_responses()
    builder.bin_responses()
    min_frequency = 1000
    max_frequency = 8000
    builder.add_stimuli(
        Spectrogram(
            scaling="density",
            min_frequency=min_frequency,
            max_frequency=max_frequency,
            log_transform_compress=1,
        )
    )
    builder.create_time_lags()
    dataset = builder.get_dataset()
    X, Y = dataset[:]
    assert X.shape[0] == Y.shape[0]
    frequency_bands = dataset._get_stimuli().columns
    assert ((frequency_bands > min_frequency) & (frequency_bands < max_frequency)).all()


def test_gammatone(real_data_source):
    builder = DatasetBuilder()
    builder.set_data_source(real_data_source)
    builder.load_responses()
    builder.bin_responses()
    frequency_bin_count = 50
    builder.add_stimuli(Gammatone(frequency_bin_count=frequency_bin_count))
    builder.create_time_lags()
    dataset = builder.get_dataset()
    X, Y = dataset[:]
    assert Y.shape[1] == frequency_bin_count
    assert X.shape[0] == Y.shape[0]


def test_syllable(real_data_source):
    builder = DatasetBuilder()
    builder.set_data_source(real_data_source)
    builder.load_responses()
    builder.bin_responses()

    builder.add_stimuli(SyllableCategorical())
    builder.create_time_lags()
    dataset = builder.get_dataset()
    X, Y = dataset[:]
    assert X.shape[0] == Y.shape[0]
    trial = dataset._get_trial_data().iloc[0]
    assert np.allclose(
        dataset._get_responses().loc[trial.name].index,
        dataset._get_stimuli().loc[trial['stimulus.name']].index,
        atol=1e-6,
    )


def test_datasource_has_diff_stimuli(stimtrial_pprox):
    sample_rate = 44100
    samples = np.random.normal(0, 1000, 44100)
    # note that the stimuli dict contains a strict superset of the stimuli referenced
    # in the responses dict. That's the point of this test
    stimuli = {"song_1": (sample_rate, samples), "song_2": (sample_rate, samples)}
    responses = stimtrial_pprox
    builder = DatasetBuilder()
    builder.set_data_source(MemorySource(responses, stimuli))
    builder.load_responses()
    builder.bin_responses()
    frequency_bin_count = 50
    builder.add_stimuli(Gammatone(frequency_bin_count=frequency_bin_count))
    builder.create_time_lags()
    dataset = builder.get_dataset()
    X, Y = dataset[:]
    assert Y.shape[1] == frequency_bin_count
    assert X.shape[0] == Y.shape[0]


@pytest.mark.parametrize("format", [Gammatone, Spectrogram])
def test_known_spectrogram(known_spectrogram_source, format):
    # make sure that the frequency band with the most power is the one that
    # is closest to 3kHz for a signal that is a sine wave oscillating at 3kHz
    builder = DatasetBuilder()
    builder.set_data_source(known_spectrogram_source)
    builder.load_responses()
    builder.add_stimuli(format(), time_step=0.005)
    dataset = builder.get_dataset()
    spec = dataset._get_stimuli().loc["song_1"]
    freq_band = 3e3
    closest_column = (spec.columns.to_series() - freq_band).abs().argmin()
    assert spec.mean().argmax() == closest_column



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['preconstruct', 'preconstruct._tests']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0',
 'appdirs>=1.4.4,<2.0.0',
 'joblib>=1.2.0,<2.0.0',
 'melizalab-gammatone>=0.1.1,<0.2.0',
 'numpy>=1.21.1,<2.0.0',
 'pandas>=1.3.1,<2.0.0',
 'parse>=1.19.0,<2.0.0',
 'scipy>=1.7.2,<2.0.0',
 'single-source>=0.3.0,<0.4.0']

setup_kwargs = {
    'name': 'preconstruct',
    'version': '0.2.3',
    'description': 'creates a dataset that is arranged to facilitate neural decoding',
    'long_description': 'None',
    'author': 'Jonah Weissman',
    'author_email': 'jw5av@virginia.edu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.12',
}


setup(**setup_kwargs)
